# -*- coding: utf-8 -*-

# @Time  : 2022/5/9 17:46
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License

"""
Abstract classes for building graph representations consist with ``pytorch-geometric``.

All the Graph in this part should return data as following:

Each Graph data (for each structure):

``x``: Node feature matrix. np.ndarray, with shape [num_nodes, nfeat_node]

``pos``: Node position matrix. np.ndarray, with shape [num_nodes, num_dimensions]

``y``: target. np.ndarray, shape (1, num_target) , default shape (1,)

``z``: atom numbers. np.ndarray, with shape [num_nodes,]

``state_attr``: state feature. np.ndarray, shape (1, nfeat_state)

``edge_index``: Graph connectivity in COO format. np.ndarray, with shape [2, num_edges] and type torch.long

``edge_attr``: Edge feature matrix. np.ndarray,  with shape [num_edges, nfeat_edge]

Where the state_attr is added newly.

"""

import os
import warnings
from shutil import rmtree
from typing import List, Iterable, Union, Tuple

import ase.data as ase_data
import numpy as np
import torch
from pymatgen.core import Structure, Element
from pymatgen.optimization.neighbors import find_points_in_spheres
from torch_geometric.data import Data

from mgetool.tool import parallelize, batch_parallelize


def data_merge(data_sub: List[Data], merge_key=None):
    if merge_key is None:
        merge_key = ["state_attr", "edge_attr", "x"]
    data_base = data_sub[0]
    for di in data_sub[1:]:
        ks = di.keys
        for k in ks:
            v = getattr(di, k)
            if v is not None:
                if k not in merge_key:
                    data_base.__setattr__(k, v)
                else:
                    if not hasattr(data_base, k):
                        data_base.__setattr__(k, v)
                    elif isinstance(getattr(data_base, k), torch.Tensor) and isinstance(v, torch.Tensor):
                        v_base = getattr(data_base, k)
                        v_base_s = len(v_base.shape)
                        v_s = len(v.shape)
                        if v_base_s == 0 or v_s == 0 or all((v_base_s == 1, v_s == 1)):
                            data_base.__setattr__(k, torch.cat([v_base.flatten(), v.flatten()]))
                        elif all((v_base_s >= 2, v_s >= 2)) and v_base.shape[0] == v_base.shape[0]:
                            data_base.__setattr__(k, torch.cat([v_base, v], dim=1))
                        else:
                            data_base.__setattr__(k, v)
                    else:
                        data_base.__setattr__(k, v)
    return data_base


def data_dtype(data: Data) -> Data:
    ks = data.keys
    for k in ks:
        v = getattr(data, k)
        if isinstance(v, torch.Tensor):
            if v.dtype == torch.float64:
                data.__setattr__(k, v.float())
    return data


class StructureToData:
    def __init__(self, sub_converters: List = None, merge_key=None, n_jobs: int = 1,
                 batch_calculate: bool = True,
                 batch_size: int = 30):
        self.n_jobs = n_jobs
        self.batch_calculate = batch_calculate
        self.batch_size = batch_size
        self.sub_converters = sub_converters
        self.merge_key = merge_key

    def __call__(self, *args, **kwargs):
        return self.convert(*args, **kwargs)

    def convert(self, data: Structure, y=None, state_attr=None, **kwargs) -> Data:
        """

        Args:
            data: (Structure)
            y: (float,int)
            state_attr: (float,int,np.ndarray)
            **kwargs: kwargs are just applied on main converter ranther than the subconverters

        Returns:
            data:(Data),pyg Data

        """
        if self.sub_converters is None:
            data = self._convert(data, y=y, state_attr=state_attr, **kwargs)
        else:
            data_sub = [ci.convert(data, y=y, state_attr=state_attr, **kwargs) for ci in self.sub_converters]
            data = self._convert(data, y=y, state_attr=state_attr, **kwargs)
            data_sub.append(data)

            data = data_merge(data_sub)
            data = data_dtype(data)

        return data

    def _wrapper(self, *args, **kwargs):
        try:
            try:
                con = self.convert(*args, **kwargs)
            except TypeError as e:
                print(e)
                raise TypeError("Please check the above errors")

            if isinstance(con, (List, Tuple)):
                if len(con) == 2 and isinstance(con[1], bool):
                    pass
                else:
                    con = (con, True)
            else:
                con = (con, True)
            return con

        except BaseException as e:
            warnings.warn(e.__str__())
            print(f"Bad conversion for:{args}, Check and index self.support_ to drop error data.", )
            return np.nan, False

    def transform(self, structures: List[Structure], **kwargs) -> List[Data]:
        """

        Args:
            structures:(list) Preprocessing of samples need to transform to Graph.
            state_attr: (list)
                preprocessing of samples need to add to Graph.
            y: (list)
                Target to train against (the same size with structure)

            **kwargs:

        Returns:
            list of graphs:
                List of dict

        """
        assert isinstance(structures, Iterable)
        if hasattr(structures, "__len__"):
            assert len(structures) > 0, "Empty input data!"

        le = len(structures)

        for i in kwargs.keys():
            if kwargs[i] is None:
                kwargs[i] = [kwargs[i]] * len(structures)
            elif not isinstance(kwargs[i], Iterable):
                kwargs[i] = [kwargs[i]] * len(structures)

        try:
            kw = [{k: v[i] for k, v in kwargs.items()} for i in range(le)]
        except IndexError as e:
            print(e)
            raise IndexError("Make sure the other parameters such as y and state_attr"
                             " are with same number (length) of structure.")

        iterables = zip(structures, kw)

        if not self.batch_calculate:
            res = parallelize(self.n_jobs, self._wrapper(), iterables, tq=True, respective=True,
                              respective_kwargs=True)

        else:
            res = batch_parallelize(self.n_jobs, self._wrapper, iterables, respective=True,
                                    respective_kwargs=True,
                                    tq=True, mode="j", batch_size=self.batch_size)

        ret, self.support_ = zip(*res)

        ret = [i for i, j in zip(ret, self.support_) if j is True]

        return ret

    def check_dup(self, structures, file_names="composition_name"):
        """Check the names duplication"""
        names = [i.composition.reduced_formula for i in structures]
        if file_names == "composition_name" and len(names) != len(set(names)):
            raise KeyError("There are same composition name for different structure, "
                           "please use file_names='rank_number' "
                           "to definition or specific names list.")
        elif len(set(file_names)) == len(structures):
            return file_names
        elif file_names == "rank_number":
            return ["raw_data_{}".format(i) for i in range(len(structures))]
        else:
            return names

    def save(self, obj, name, root_dir=".") -> None:
        """Save."""
        torch.save(obj, os.path.join(root_dir, "raw", '{}.pt'.format(name)))

    def transform_and_save(self, data, y=None, state_attr=None, root_dir=".",
                           file_names="composition_name", save_mode="o", **kwargs):
        r"""Save the data to 'root_dir/raw' if save_mode="i", else 'root_dir', compact with InMemoryDatasetGeo"""
        raw_path = os.path.join(root_dir, "raw")
        if os.path.isdir(raw_path):
            rmtree(raw_path)
        os.makedirs(raw_path)

        result = self.transform(data, y=y, state_attr=state_attr, **kwargs)

        print("Save raw files to {}.".format(raw_path))
        if save_mode in ["i", "r", "respective"]:
            fns = self.check_dup(data, file_names=file_names)
            [self.save(i, j, root_dir) for i, j in zip(result, fns)]
        else:
            self.save(result, "raw_data", root_dir=root_dir)
        print("Done.")
        return result

    @staticmethod
    def _convert(data: Structure, y=None, state_attr=None, **kwargs) -> Data:
        z = torch.tensor(list(data.atomic_numbers))
        pos = torch.from_numpy(data.cart_coords)
        if state_attr is not None:
            state_attr = torch.tensor(state_attr, dtype=torch.float32)
            state_attr = state_attr.reshape(1, 1) if len(state_attr.shape) == 0 else state_attr
        if y is not None:
            y = torch.tensor(y).reshape(1, 1)

        for key, value in kwargs.items():
            try:
                value = torch.tensor(value)
                value = value.reshape(1, 1) if len(value.shape) == 0 else value
                kwargs[key] = value
            except ValueError:
                pass

        return Data(y=y, pos=pos, z=z, state_attr=state_attr, **kwargs)


class AddSAPymatgen(StructureToData):
    """
    Add state_attr.
    pymatgen's structure.num_sites, structure.ntypesp, structure.density, structure.volume"""

    def __init__(self, *args, prop_name=None, **kwargs):
        if prop_name is None:
            self.prop_name = ['num_sites', 'ntypesp', 'density', 'volume']
        else:
            self.prop_name = prop_name
        super(AddSAPymatgen, self).__init__(*args, **kwargs)

    @staticmethod
    def _convert(data: Structure, **kwargs) -> Data:
        state_attr = torch.tensor([data.num_sites, data.ntypesp, data.density, data.volume])
        return Data(state_attr=state_attr)


class AddXPymatgen(StructureToData):
    """
    Add x.

    Get pymatgen element preprocessing.
    prop_name = [
    "atomic_radius",
    "atomic_mass",
    "number",
    "max_oxidation_state",
    "min_oxidation_state",
    "row",
    "group",
    "atomic_radius_calculated",
    "mendeleev_no",
    "average_ionic_radius",
    "average_cationic_radius",
    "average_anionic_radius",]
    """

    def __init__(self, *args, prop_name=None, **kwargs):
        super(AddXPymatgen, self).__init__(*args, **kwargs)
        if prop_name is None:
            self.prop_name = ["atomic_mass", "average_ionic_radius", "average_anionic_radius",
                              "atomic_radius_calculated"]
        else:
            self.prop_name = prop_name
        self.da = [Element.from_Z(i) for i in range(1, 119)]
        self.da.insert(0, self.da[0])  # for start from 1

    def _convert(self, data: Structure, **kwargs) -> Data:
        x = torch.tensor([[getattr(self.da[i], pi) for pi in self.prop_name] for i in data.atomic_numbers])
        return Data(x=x)


class AddXASE(StructureToData):
    """
    Add x.

    Get pymatgen element preprocessing.
    prop_name = [
    'atomic_masses', 'covalent_radii'
    ]
    """

    def __init__(self, *args, prop_name=None, **kwargs):
        super(AddXASE, self).__init__(*args, **kwargs)
        if prop_name is None:
            self.prop_name = ['atomic_masses', 'covalent_radii']
        else:
            self.prop_name = prop_name
        arrays = np.concatenate([getattr(ase_data, i).reshape(-1, 1) for i in self.prop_name], axis=1)
        self.arrays = torch.from_numpy(arrays)

    def _convert(self, data: Structure, **kwargs) -> Data:
        x = self.arrays[data.atomic_numbers, :]
        return Data(x=x)


class AddXArray(StructureToData):
    """
    Add x by np.array (2D).
    The array are insert in 0 position automatically in code.

    """

    def __init__(self, array: np.ndarray, *args, **kwargs):
        super(AddXArray, self).__init__(*args, **kwargs)
        if array.ndim == 1:
            array = array.reshape(-1, 1)
        array = np.concatenate((array[0, :].reshape(1, -1), array), axis=0)  # add one line in 0.
        self.arrays = torch.from_numpy(array)

    def _convert(self, data: Structure, **kwargs) -> Data:
        x = self.arrays[data.atomic_numbers, :]
        return Data(x=x)


def _re_pbc(pbc: Union[bool, List[bool], np.ndarray], return_type="bool"):
    if pbc is True:
        pbc = [1, 1, 1]
    elif pbc is False:
        pbc = [0, 0, 0]
    elif isinstance(pbc, Iterable):
        pbc = [1 if i is True or i == 1 else 0 for i in pbc]
    else:
        raise TypeError("Can't accept {}".format(pbc))
    if return_type == "bool":
        pbc = np.array(pbc) == 1
    else:
        pbc = np.array(pbc)
    return pbc


def _get_r_in_spheres(data, pbc=True, cutoff=5.0, numerical_tol=1e-6):
    if isinstance(data, Structure):
        lattice_matrix = np.ascontiguousarray(np.array(data.lattice.matrix), dtype=float)
        if pbc is not False:
            pbc = _re_pbc(pbc, return_type="int")
        else:
            pbc = np.array([0, 0, 0])
    else:
        raise ValueError("structure type not supported")

    r = float(cutoff)
    cart_coords = np.ascontiguousarray(np.array(data.cart_coords), dtype=float)
    center_indices, neighbor_indices, images, distances = find_points_in_spheres(
        cart_coords, cart_coords, r=r, pbc=pbc, lattice=lattice_matrix, tol=numerical_tol
    )
    center_indices = center_indices.astype(np.int64)
    neighbor_indices = neighbor_indices.astype(np.int64)
    # images = images.astype(np.int64)
    distances = distances.astype(np.float32)
    exclude_self = (distances > numerical_tol)
    # exclude_self = (center_indices != neighbor_indices) | (distances > numerical_tol)
    center_indices = center_indices[exclude_self]
    neighbor_indices = neighbor_indices[exclude_self]
    distances = distances[exclude_self]
    return center_indices, neighbor_indices, distances, None


def _get_xyz_in_spheres(data, pbc=True, cutoff=5.0, numerical_tol=1e-6):
    if isinstance(data, Structure):
        lattice_matrix = np.ascontiguousarray(np.array(data.lattice.matrix), dtype=float)
        if pbc is not False:
            pbc = _re_pbc(pbc, return_type="int")
        else:
            pbc = np.array([0, 0, 0])
    else:
        raise ValueError("structure type not supported")
    r = float(cutoff)
    cart_coords = np.ascontiguousarray(np.array(data.cart_coords), dtype=float)
    center_indices, neighbor_indices, images, distances = find_points_in_spheres(
        cart_coords, cart_coords, r=r, pbc=pbc, lattice=lattice_matrix, tol=numerical_tol
    )
    center_indices = center_indices.astype(np.int64)
    neighbor_indices = neighbor_indices.astype(np.int64)
    images = images.astype(np.int64)
    distances = distances.astype(np.float32)
    exclude_self = (distances > numerical_tol)

    center_indices = center_indices[exclude_self]
    neighbor_indices = neighbor_indices[exclude_self]
    distances = distances[exclude_self]
    images = images[exclude_self]

    offset = np.dot(images, data.lattice.matrix)

    coords = cart_coords[neighbor_indices, :] - cart_coords[center_indices, :]
    xyz = offset + coords

    return center_indices, neighbor_indices, distances, xyz


class AddPBCEdgeDistance(StructureToData):
    """
    Add Edge with PBC.
    make sure

    Get pymatgen element preprocessing.
    prop_name = [
    'vdw_radii', 'reference_states', 'atomic_masses', 'covalent_radii'
    ]
    """

    def __init__(self, *args, cutoff: float = 5.0,
                 numerical_tol: float = 1e-6,
                 pbc=True, **kwargs):
        """
        Get graph representations from structure within cutoff.

        Args:
            structure (pymatgen Structure or molecule)
            cutoff (float): cutoff radius
            numerical_tol (float): numerical tolerance
            nn_strategy(str,None):not used
        Returns:
            center_indices, neighbor_indices, distances
        """
        super(AddPBCEdgeDistance, self).__init__(*args, **kwargs)
        self.numerical_tol = numerical_tol
        self.pbc = pbc
        self.cutoff = cutoff

    def _convert(self, data: Structure, **kwargs) -> Data:
        center_indices, neighbor_indices, distances, _ = _get_r_in_spheres(data, pbc=self.pbc,
                                                                           cutoff=self.cutoff,
                                                                           numerical_tol=self.numerical_tol)

        assert len(set(center_indices)) == data.num_sites, f"Some atom are independent at cutoff: {self.cutoff}"

        edge_index = torch.vstack([torch.from_numpy(center_indices),
                                   torch.from_numpy(neighbor_indices)])
        edge_weight = torch.from_numpy(distances)

        return Data(edge_index=edge_index, edge_weight=edge_weight)


class AddPBCEdgeXYZ(StructureToData):
    """
    Add edge xyz.
    make sure.

    Get pymatgen element preprocessing.
    prop_name = [
    'vdw_radii', 'reference_states', 'atomic_masses', 'covalent_radii'
    ]
    """

    def __init__(self, *args, cutoff: float = 5.0,
                 numerical_tol: float = 1e-6,
                 pbc=True, **kwargs):
        """
        Get graph representations from structure within cutoff.

        Args:
            structure (pymatgen Structure or molecule)
            cutoff (float): cutoff radius
            numerical_tol (float): numerical tolerance
            nn_strategy(str,None):not used
        Returns:
            center_indices, neighbor_indices, distances
        """
        super(AddPBCEdgeXYZ, self).__init__(*args, **kwargs)
        self.numerical_tol = numerical_tol
        self.pbc = pbc
        self.cutoff = cutoff

    def _convert(self, data: Structure, **kwargs) -> Data:
        center_indices, neighbor_indices, distances, xyz = _get_xyz_in_spheres(data, pbc=self.pbc,
                                                                               cutoff=self.cutoff,
                                                                               numerical_tol=self.numerical_tol)

        assert len(set(center_indices)) == data.num_sites, f"Some atom are independent at cutoff: {self.cutoff}"

        assert center_indices.shape == neighbor_indices.shape
        edge_index = torch.vstack([torch.from_numpy(center_indices),
                                   torch.from_numpy(neighbor_indices)])
        edge_weight = torch.from_numpy(distances)
        edge_attr = torch.from_numpy(xyz)

        return Data(edge_index=edge_index, edge_weight=edge_weight, edge_attr=edge_attr)



import unittest

import numpy as np
from pymatgen.core import Structure

from pyg_extension.data.structuretodata import AddSAPymatgen, AddXPymatgen, AddXASE, \
    StructureToData, AddXArray, \
    AddPBCEdgeXYZ

data = Structure.from_file("POSCAR")
datas = [data] * 10000


class MyTestCase(unittest.TestCase):
    def test_AddSAPymatgen(self):
        addsap = AddSAPymatgen()
        res = addsap.convert(data, y=None, state_attr=None)
        print(res)

    def test_AddSAPymatgen2(self):
        sub_converters = AddXPymatgen()
        addsap = AddSAPymatgen(sub_converters=[sub_converters])
        res = addsap.convert(data, y=None, state_attr=None)
        print(res)

    def test_AddXASE(self):
        sub_converters1 = AddSAPymatgen()
        sub_converters2 = AddXPymatgen()
        addsap = StructureToData(sub_converters=[sub_converters1, sub_converters2], n_jobs=3)
        # res = addsap.transform(datas, y=None, )
        res = addsap.transform(datas, y=None, state_attr=[[3, 4, 5]] * 10000)
        print(res)

    def test_AddXArray(self):
        array = np.random.random((120, 10))
        sub_converters1 = AddXASE()
        sub_converters2 = AddXPymatgen()
        sub_converters3 = AddXArray(array)
        sub_converters4 = AddSAPymatgen()
        addsap = StructureToData(sub_converters=[sub_converters1, sub_converters2, sub_converters3, sub_converters4],
                                 n_jobs=1)
        # res = addsap.transform(datas, y=None, )
        res = addsap.transform(datas, y=None)
        print(res)

    def test_AddPBCEdgeDistance(self):
        addsap = StructureToData(sub_converters=[AddXPymatgen(), AddSAPymatgen(), AddPBCEdgeXYZ()], n_jobs=4)
        res = addsap.transform_and_save(datas)
        print(res)


if __name__ == '__main__':
    unittest.main()



import unittest

import torch
from sklearn.preprocessing import MinMaxScaler
from sklearn.utils import shuffle
from torch import nn, Tensor
from torch.nn import Module, ReLU, Sequential
from torch.utils.data import DataLoader, TensorDataset

from mgetool.show import BasePlot
from pyg_extension.nn.resnet import ResBlockSameSize, ResBlockDiffSize


class Li2(nn.Linear):
    def forward(self, input: Tensor, **kwargs) -> Tensor:
        return super().forward(input)


class ResNet(Module):
    def __init__(self, fea):
        super().__init__()

        self.block1 = ResBlockDiffSize(Li2, layer_size_seq=(fea, 16, 32, 64), force_x_match=True)
        self.block2 = ResBlockSameSize(Li2, n_res=1, in_features=64, out_features=64, force_x_match=True)
        self.out = Sequential(Li2(64, 128), ReLU(), Li2(128, 1))

    def forward(self, x):
        x = self.block1(x)
        x = self.block2(x)
        x = self.out(x)
        return x


class MyTestCase(unittest.TestCase):
    def setUp(self) -> None:
        from sklearn.datasets import fetch_california_housing
        X, y = fetch_california_housing(return_X_y=True)
        X, y = shuffle(X, y, random_state=0)
        mms = MinMaxScaler()

        X = X[:5000]
        y = y[:5000]
        X = mms.fit_transform(X)
        X = torch.from_numpy(X).float()
        y = torch.from_numpy(y).float().reshape(-1, 1)
        self.fea = X.shape[1]
        self.X = X
        self.y = y

    def test_something(self):
        resn = ResNet(self.fea)
        yb = resn(self.X)
        assert yb.shape == self.y.shape

    def test_flow(self):

        model = ResNet(self.fea)
        me = "cuda:0"
        me = "cpu"
        if me == "cuda:0":
            model.to("cuda:0")
            X = self.X.cuda()
            y = self.y.cuda()
        else:
            X = self.X
            y = self.y
        td = TensorDataset(X, y)

        train_loader = DataLoader(td, batch_size=256)
        optimizer = torch.optim.Adam(model.parameters(), lr=0.005)
        model.train()
        loss_method = torch.nn.L1Loss()

        for epi in range(500):
            lossi = 0
            for bx, by in train_loader:
                model.zero_grad()

                y_pred = model(bx)

                lossi = loss_method(y_pred, by)
                lossi.backward()
                optimizer.step()
            print(lossi)

        bp = BasePlot()
        plt = bp.scatter_45_line(by.detach().numpy(), y_pred.detach().numpy(), strx='y_true', stry='y_predict')
        plt.show()


if __name__ == '__main__':
    unittest.main()



import math

import torch
from torch_cluster import radius_graph
from torch_geometric.transforms import BaseTransform
from torch_geometric.utils import remove_self_loops


class AddEdge(BaseTransform):
    """Add edge index, with no pbc."""

    def __init__(self, cutoff=5.0, ):
        self.cutoff = cutoff

    def __call__(self, data):
        edge_index = radius_graph(data.pos, r=self.cutoff, batch=data.batch)
        edge_index, _ = remove_self_loops(edge_index)
        row, col = edge_index[0], edge_index[1]
        edge_weight = (data.pos[row] - data.pos[col]).norm(dim=-1)

        data.edge_index = edge_index
        data.edge_weight = edge_weight
        data.edge_attr = edge_weight.reshape(-1, 1)
        return data


class AddAttrToWeight(BaseTransform):
    """Add edge index, with no pbc."""

    def __call__(self, data):
        if not hasattr(data, "edge_weight") or data.edge_weight is None:
            data.edge_weight = torch.linalg.norm(data.edge_attr, dim=1)
        return data


class AddAttrSumToAttrAndWeight(BaseTransform):
    """Add edge index, with no pbc."""

    def __call__(self, data):
        if not hasattr(data, "edge_weight") or data.edge_weight is None:
            data.edge_weight = torch.linalg.norm(data.edge_attr, dim=1)
            data.edge_attr = data.edge_weight.reshape(-1, 1)
        return data


class AddAttrSumToAttr(BaseTransform):
    """Add edge index, with no pbc."""

    def __call__(self, data):
        assert hasattr(data, "edge_attr") and data.edge_attr is not None
        assert data.edge_attr.shape[1] == 3
        if not hasattr(data, "edge_weight") or data.edge_weight is None:
            data.edge_weight = torch.linalg.norm(data.edge_attr, dim=1)

        wei = data.edge_weight

        data.edge_attr = torch.cat((wei.reshape(-1, 1), data.edge_attr), dim=1)

        return data


class DistributeEdgeAttr(BaseTransform):
    """Compact with rotnet network (deep potential)"""

    def __init__(self, r_cs=2.0, r_c=6.0, cat_weight_attr=True):
        super().__init__()
        self.r_cs = r_cs
        self.r_c = r_c
        self.cat_weight_attr = cat_weight_attr

    def __call__(self, data):
        assert hasattr(data, "edge_attr") and data.edge_attr is not None
        assert data.edge_attr.shape[1] == 3
        if not hasattr(data, "edge_weight") or data.edge_weight is None:
            data.edge_weight = torch.linalg.norm(data.edge_attr, dim=1)

        wei = data.edge_weight

        attr = torch.cat((wei.reshape(-1, 1), data.edge_attr), dim=1)

        r_cs = self.r_cs
        r_c = self.r_c

        sr = torch.clone(attr[:, 0])
        r_m1 = sr <= r_cs
        r_m3 = sr >= r_c
        r_m2 = ~(r_m3 | r_m1)

        sr[r_m1] = 1 / sr[r_m1]

        u = (sr[r_m2] - r_cs) / (self.r_c - self.r_cs)
        sr[r_m2] = 1 / sr[r_m2] * (0.5 * torch.cos(math.pi * u) + 0.5)  # from deep Potential
        sr[r_m3] = 0

        attr = attr / attr[:, 0].reshape(-1, 1) * sr.reshape(-1, 1)

        data.edge_attr = attr
        data.edge_weight = sr
        return data



"""
Operation utilities on lists and arrays, for torch,

Notes
    just for network.
"""

from collections import abc
from functools import wraps
from typing import Union, List, Sequence, Optional

import numpy as np
import torch
from sklearn.model_selection import train_test_split
from torch import Tensor
from torch.nn import Module
from torch_geometric.nn.conv.utils.helpers import expand_left
from torch_scatter import gather_csr
from torch_sparse import SparseTensor


def to_list(x: Union[abc.Iterable, np.ndarray]) -> List:
    """
    If x is not a list, convert it to list
    """
    if isinstance(x, abc.Iterable):
        return list(x)
    if isinstance(x, np.ndarray):
        return x.tolist()  # noqa
    return [x]


def fast_label_binarize(value: List, labels: List) -> List[int]:
    """Faster version of label binarize

    `label_binarize` from scikit-learn is slow when run 1 label at a time.
    `label_binarize` also is efficient for large numbers of classes, which is not
    common in `megnet`

    Args:
        value: Value to encode
        labels (list): Possible class values
    Returns:
        ([int]): List of integers
    """

    if len(labels) == 2:
        return [int(value == labels[0])]
    output = [0] * len(labels)
    if value in labels:
        output[labels.index(value)] = 1
    return output


def check_shape(array: Optional[np.ndarray], shape: Sequence) -> bool:
    """
    Check if array complies with shape. Shape is a sequence of
    integer that may end with None. If None is at the end of shape,
    then any shapes in array after that dimension will match with shape.

    Example: array with shape [10, 20, 30, 40] matches with [10, 20, None], but
        does not match with shape [10, 20, 30, 20]

    Args:
        array (np.ndarray or None): array to be checked
        shape (Sequence): integer array shape, it may ends with None
    Returns: bool
    """
    if array is None:
        return True
    if all(i is None for i in shape):
        return True

    array_shape = array.shape
    valid_dims = [i for i in shape if i is not None]
    n_for_check = len(valid_dims)
    return all(i == j for i, j in zip(array_shape[:n_for_check], valid_dims))


def train_test_pack(*arrays, out=0, **options):
    """Split arrays or matrices into random train and test subsets

    Quick utility that wraps input validation and
    ``next(ShuffleSplit().split(X, y))`` and application to input data
    into a single call for splitting (and optionally subsampling) data in a
    oneliner.

    Parameters
    ----------
    *arrays : sequence of indexables with same length / shape[0]
        Allowed inputs are lists, numpy arrays, scipy-sparse
        matrices or pandas dataframes.

    test_size : float or int, default=None
        If float, should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split. If int, represents the
        absolute number of test samples. If None, the value is set to the
        complement of the train size. If ``train_size`` is also None, it will
        be set to 0.25.

    train_size : float or int, default=None
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the train split. If
        int, represents the absolute number of train samples. If None,
        the value is automatically set to the complement of the test size.

    random_state : int or RandomState instance, default=None
        Controls the shuffling applied to the data before applying the split.
        Pass an int for reproducible output across multiple function calls.

    shuffle : bool, default=True
        Whether or not to shuffle the data before splitting. If shuffle=False
        then stratify must be None.

    stratify : array-like, default=None
        If not None, data is split in a stratified fashion, using this as
        the class labels.

    out:int
        pack the pack to format,
        pack=None: (0:n-1)
        pack=0: (0:n-1:2, 1:n:2)
        pack=1: (0:n-3:2, -2, 1:n-2:2, -1)
        pack=2: (0:n-5:2, -4, -2, 1:n-4:2, -3, -1)

    Examples
    --------
    >>> X_train, y_train, X_test, y_test = train_test_split(
    ...     *X, y, test_size=0.33, random_state=42)

    """
    num = len(arrays)
    assert num >= out

    train_test_data = train_test_split(*arrays, **options)
    le = len(train_test_data)
    if out == 0:
        X_train = [train_test_data[i] for i in range(le) if i % 2 == 0]
        X_test = [train_test_data[i] for i in range(le) if i % 2 == 1]
        if num == 1:
            X_train, X_test = X_train[0], X_test[0]
        return X_train, X_test
    elif out == 1:
        (*X_train, y_train) = [train_test_data[i] for i in range(le) if i % 2 == 0]
        (*X_test, y_test) = [train_test_data[i] for i in range(le) if i % 2 == 1]
        if num == 2:
            X_train, X_test = X_train[0], X_test[0]
        return X_train, y_train, X_test, y_test
    elif out == 2:
        (*X_train, y_train, label_train) = [train_test_data[i] for i in range(le) if i % 2 == 0]
        (*X_test, y_test, label_test) = [train_test_data[i] for i in range(le) if i % 2 == 1]
        if num == 3:
            X_train, X_test = X_train[0], X_test[0]
        return X_train, y_train, X_test, y_test, label_train, label_test
    else:
        return train_test_data


def getter_arr(obj, pi):
    """Get prop.
    """
    if "." in pi:
        pis = list(pi.split("."))
        pis.reverse()
        while len(pis):
            s = pis.pop()
            obj = getter_arr(obj, s)
        return obj
    elif "()" in pi:
        return getattr(obj, pi[:-2])()

    else:
        return getattr(obj, pi)


def temp_jump(mark=0, temp_device=None, old_device=None, temp=True, back=True):
    def f(func):
        return _temp_jump(func, mark=mark, temp_device=temp_device, old_device=old_device, temp=temp, back=back)

    return f


def temp_jump_cpu(mark=0, temp_device="cpu", old_device=None, temp=True, back=True):
    def f(func):
        return _temp_jump(func, mark=mark, temp_device=temp_device, old_device=old_device, temp=temp, back=back)

    return f


def _temp_jump(func, mark=0, temp_device="cpu", old_device=None, temp=True, back=True):
    """temp to cpu to calculate and re-back the init device data."""

    @wraps(func)
    def wrapper(*args, **kwargs):

        if temp_device is None:
            device = args[0].device
        else:
            device = torch.device(temp_device) if isinstance(temp_device, str) else temp_device

        if old_device is None:
            device2 = args[mark + 1].device if len(args) > 1 else list(kwargs.values())[0].device
        else:
            device2 = torch.device(old_device) if isinstance(old_device, str) else old_device

        if temp:
            args2 = [args[0]]
            for i in args[1:]:
                try:
                    args2.append(i.to(device=device, copy=False))
                except AttributeError:
                    args2.append(i)
            kwargs2 = {}
            for k, v in kwargs.items():
                try:
                    if isinstance(v, tuple):
                        kwargs2[k] = [i.to(device=device, copy=False) for i in v]
                    else:
                        kwargs2[k] = v.to(device=device, copy=False)
                except AttributeError:
                    kwargs2[k] = v

            result = func(*args2, **kwargs2)
        else:
            result = func(*args, **kwargs)

        if back:
            if isinstance(result, tuple):
                result2 = []
                for i in result:
                    try:
                        result2.append(i.to(device=device2, copy=False))
                    except AttributeError:
                        result2.append(i)
            else:
                try:
                    result2 = result.to(device=device2, copy=False)
                except AttributeError:
                    result2 = result
            return result2
        else:
            return result

    return wrapper


def check_device(mode: Module):
    device = _check_device(mode)
    return torch.device("cpu") if device is None else device


def _check_device(mode: Module):
    device = None
    for i in mode.children():
        if hasattr(i, "weight"):
            device = i.weight.device
            break
        elif hasattr(i, "bias"):
            device = i.bias.device
            break
        elif len(list(i.children())) > 0:
            device = check_device(i)
            if device is not None:
                break
    return device


#
def sparse_eg_single(edge_index, num_nodes, num_edges, edge_weight=None, **kwargs):
    """
    From:

    import torch_geometric.transforms as T
    T.ToSparseTensor()

    Make sure all the edge prop is spared together!!!
    for example edge_index,edge_weight,edge_attr...,due to index is resort"""

    (row, col), N, E = edge_index, num_nodes, num_edges
    perm = (col * N + row).argsort()
    row, col = row[perm], col[perm]

    if edge_weight:
        value = edge_weight[perm]
    else:
        value = None

    others = [value, ]
    for key, item in kwargs:
        if item.size(0) == E:
            others.append(item[perm])

    adj_t = SparseTensor(row=col, col=row, value=value,
                         sparse_sizes=(N, N), is_sorted=True)

    return adj_t, others


def get_ptr(index):
    """Trans batch index to ptr"""
    return torch.ops.torch_sparse.ind2ptr(index, index[-1] + 1)


def lift_jump_index_select(self, src, edge_index, dim):
    """From MessagePassing.__lift__, jump edge_attr from adj"""
    if isinstance(edge_index, Tensor):
        index = edge_index[dim]
        return src.index_select(self.node_dim, index)
    elif isinstance(edge_index, SparseTensor):
        if dim == 1:
            rowptr = edge_index.storage.rowptr()
            rowptr = expand_left(rowptr, dim=self.node_dim, dims=src.dim())
            return gather_csr(src, rowptr)
        elif dim == 0:
            col = edge_index.storage.col()
            # return src.index_select(self.node_dim, col)
            return src[col]
    raise ValueError



import copy
import os
import os.path as osp
from shutil import rmtree
from typing import List, Callable, Dict, Union

import torch
from torch.utils.data import Dataset as tD
from torch_geometric.data import Data, Dataset
from torch_geometric.data import InMemoryDataset


class SimpleDataset(tD):
    """ Data list with shuffle, for small data.
    For small data <= 500.
    Load data from Memory, and stored in Memory for use.

    Examples
    -----------

    >>> #Not with Dataset
    >>> # with data.edge_index shape (2,num_edge)
    >>> dataset = [Data.from_dict(di) for di in data]
    >>> # SparseTensor: with data.edge_index shape (num_node,num_node)
    >>> import torch_geometric.transforms as T
    >>> dataset2 = [T.ToSparseTensor(Data.from_dict(di))) for di in data]

    >>> # with SimpleDataset
    >>> # with data.edge_index shape (2,num_edge)
    >>> dataset = SimpleDataset(data)
    >>> # SparseTensor: with data.edge_index shape (num_node,num_node)
    >>> import torch_geometric.transforms as T
    >>> dataset2 = SimpleDataset(data,pre_transform=T.ToSparseTensor())

    """

    def __init__(self, data: Union[List[Data], List[Dict]], pre_filter=None, pre_transform=None,
                 transform: Callable = None):
        """

        Args:
            transform (callable, optional): A function/transform that takes in an
                :obj:`torch_geometric.data.Data` object and returns a transformed
                version. The data object will be transformed before every access.
                (default: :obj:`None`)
            pre_transform (callable, optional): A function/transform that takes in
                an :obj:`torch_geometric.data.Data` object and returns a
                transformed version. The data object will be transformed before
                being saved to disk. (default: :obj:`None`)
            pre_filter (callable, optional): A function that takes in an
                :obj:`torch_geometric.data.Data` object and returns a boolean
                value, indicating whether the data object should be included in the
                final dataset. (default: :obj:`None`)

        """
        super(SimpleDataset, self).__init__()
        self.transform = transform
        self.pre_filter = pre_filter
        self.pre_transform = pre_transform
        if isinstance(data[0], dict):
            self.data = [Data.from_dict(di) for di in data]
        else:
            self.data = data
        self.__indices__ = None
        self.process()

    def len(self):
        return len(self.data)

    def get(self, idx):
        return self.data[idx]

    def indices(self):
        if self.__indices__ is not None:
            return self.__indices__
        else:
            return range(len(self))

    @property
    def nfeat_node(self):
        r"""Returns the number of features per node in the dataset."""
        return self[0].nfeat_node

    @property
    def num_features(self):
        r"""Alias for :py:attr:`~nfeat_node`."""
        return self.nfeat_node

    @property
    def nfeat_edge(self):
        r"""Returns the number of features per edge in the dataset."""
        return self[0].nfeat_edge

    def __len__(self):
        r"""The number of examples in the dataset."""
        if self.__indices__ is not None:
            return len(self.__indices__)
        return self.len()

    def __getitem__(self, idx):
        r"""Gets the data object at index :obj:`idx` and transforms it (in case
        a :obj:`self.transform` is given).
        In case :obj:`idx` is a slicing object, *e.g.*, :obj:`[2:5]`, a list, a
        tuple, a  LongTensor or a BoolTensor, will return a subset of the
        dataset at the specified indices."""
        if isinstance(idx, int):
            data = self.get(self.indices()[idx])
            data = data if self.transform is None else self.transform(data)
            return data
        else:
            return self.index_select(idx)

    def index_select(self, idx):
        indices = self.indices()

        if isinstance(idx, slice):
            indices = indices[idx]
        elif torch.is_tensor(idx):
            if idx.dtype == torch.long:
                if len(idx.shape) == 0:
                    idx = idx.unsqueeze(0)
                return self.index_select(idx.tolist())
            elif idx.dtype == torch.bool or idx.dtype == torch.uint8:
                return self.index_select(
                    idx.nonzero(as_tuple=False).flatten().tolist())
        elif isinstance(idx, list) or isinstance(idx, tuple):
            indices = [indices[i] for i in idx]
        else:
            raise IndexError(
                'Only integers, slices (`:`), list, tuples, and long or bool '
                'tensors are valid indices (got {}).'.format(
                    type(idx).__name__))

        dataset = copy.copy(self)
        dataset.__indices__ = indices
        return dataset

    def shuffle(self, return_perm=False):
        r"""Randomly shuffles the examples in the dataset.

        Args:
            return_perm (bool, optional): If set to :obj:`True`, will
                additionally return the random permutation used to shuffle the
                dataset. (default: :obj:`False`)
        """
        perm = torch.randperm(len(self))
        dataset = self.index_select(perm)
        return (dataset, perm) if return_perm is True else dataset

    def __repr__(self):  # pragma: no cover
        return f'{self.__class__.__name__}({len(self)})'

    def process(self):
        # Read data into huge `Data` list.
        data_list = self.data

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        self.data = data_list


class InMemoryDatasetGeo(InMemoryDataset):
    """For small data <= 2000.
    Load data from local disk, and stored in Memory for use.

    Examples
    -----------
    >>> # with data.edge_index shape (2,num_edge)
    >>> dataset = InMemoryDatasetGeo(root=".")
    >>> # SparseTensor: with data.edge_index shape (num_node,num_node)
    >>> import torch_geometric.transforms as T
    >>> dataset2 = InMemoryDatasetGeo(root=".", pre_transform=T.ToSparseTensor())

    """

    def __init__(self, root, pre_transform=None, pre_filter=None, re_process_init=True, transform=None, load_mode="o"):
        """

        Args:
            load_mode (str): load from the independent data in different files: "i",
                load from the batch data in one overall file: "o".
            re_process_init (bool): process raw data or not. if there is no raw data but processed data is offered.
                this parameter could be False. Because the default
            root (string, optional): Root directory where the dataset should be
                saved. (optional: :obj:`None`)
            transform (callable, optional): A function/transform that takes in an
                :obj:`torch_geometric.data.Data` object and returns a transformed
                version. The data object will be transformed before every access.
                (default: :obj:`None`)
            pre_transform (callable, optional): A function/transform that takes in
                an :obj:`torch_geometric.data.Data` object and returns a
                transformed version. The data object will be transformed before
                being saved to disk. (default: :obj:`None`)
            pre_filter (callable, optional): A function that takes in an
                :obj:`torch_geometric.data.Data` object and returns a boolean
                value, indicating whether the data object should be included in the
                final dataset. (default: :obj:`None`)

        """

        if load_mode in ["o", "O", "overall", "Overall"]:
            self._load = self._load_overall
            # "overall"
        elif load_mode in ["I", "R", "i", "r", "Respective", "respective"]:
            # "Respectively"
            self._load = self._load_respective
        else:
            raise NotImplementedError("load_mode = 'o' or 'i'")

        super(InMemoryDatasetGeo, self).__init__(root, transform, pre_transform, pre_filter=pre_filter)

        if re_process_init:
            self.re_process()
        else:
            self.data, self.slices = torch.load(osp.join(self.processed_paths[0]))

    def _load_overall(self):
        assert len(self.raw_paths) == 1, "There is more than one .pt file,and not sure which one to import."
        return torch.load(self.raw_paths[0])

    def _load_respective(self):
        return [torch.load(i) for i in self.raw_paths]

    @property
    def raw_file_names(self):
        return os.listdir(self.raw_dir)  # must with 'raw' dir.

    @property
    def processed_file_names(self):
        return ['data.pt']

    def process(self):
        # Read data into huge `Data` list.
        data_list = self._load()

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        self.data, self.slices = self.collate(data_list)
        torch.save((self.data, self.slices), self.processed_paths[0])

    def re_process(self):
        """Re-process for skip reset in 56 line in ``InMemoryDataset``
        `self.data, self.slices = None, None` ."""
        try:
            rmtree(self.processed_dir)
        except FileNotFoundError:
            pass
        os.makedirs(self.processed_dir)
        self.process()

        print('Done!')


class DatasetGeo(Dataset):
    """For very very huge data.
    load data from local disk each epoth, and stored in Memory temporary.

    Examples
    -----------
    >>> # with data.edge_index shape (2,num_edge)
    >>> dataset = DatasetGeo(root=".")
    >>> # SparseTensor: with data.edge_index shape (num_node,num_node)
    >>> import torch_geometric.transforms as T
    >>> dataset2 = DatasetGeo(root=".", pre_transform=T.ToSparseTensor())

    """

    def __init__(self, root, transform=None, pre_transform=None, pre_filter=None, re_process_init=True):
        """
        "Just accept mode 'i' for ``DatasetGEO``."

        Args:
            re_process_init (bool): process raw data or not. if there is no raw data but processed data is offered.
                this parameter could be False.
            root (string, optional): Root directory where the dataset should be
                saved. (optional: :obj:`None`)
            transform (callable, optional): A function/transform that takes in an
                :obj:`torch_geometric.data.Data` object and returns a transformed
                version. The data object will be transformed before every access.
                (default: :obj:`None`)
            pre_transform (callable, optional): A function/transform that takes in
                an :obj:`torch_geometric.data.Data` object and returns a
                transformed version. The data object will be transformed before
                being saved to disk. (default: :obj:`None`)
            pre_filter (callable, optional): A function that takes in an
                :obj:`torch_geometric.data.Data` object and returns a boolean
                value, indicating whether the data object should be included in the
                final dataset. (default: :obj:`None`)


        """
        super().__init__(root, transform, pre_transform, pre_filter)
        self.re_process_init = re_process_init

        super(DatasetGeo, self).__init__(root, transform, pre_transform, pre_filter=pre_filter)

        if re_process_init:
            self.re_process()
        else:
            pass

    @property
    def raw_file_names(self):
        return os.listdir(self.raw_dir)

    @staticmethod
    def _files_exist(path):
        if os.path.isdir(path):
            files = os.listdir(path)
            return len(files) > 0
        else:
            return False

    @property
    def processed_file_names(self):
        if self._files_exist(self.processed_dir) and not self.re_process_init:
            return os.listdir(self.processed_dir)
        else:
            return ["data_{}.pt".format(i) for i in range(len(self.raw_file_names))]

    def re_process(self):
        """For temporary debug"""
        rmtree(self.processed_dir)  # remove the old
        os.makedirs(self.processed_dir)
        self.process()

        print('Done!')

    def process(self):
        i = 0
        for raw_path in self.raw_paths:
            # Read data from `raw_path`.
            try:
                data = torch.load(raw_path)
            except AttributeError as e:
                print(e)
                raise AttributeError("Just accept mode 'i' for ``DatasetGEO``, which load one at a time.",
                                     "The {} may be a batch of data. "
                                     "That is , if your raw data are save in one file overall, with 'o' mode."
                                     "please turn to ``InMemoryDatasetGeo``".format(raw_path))

            if self.pre_filter is not None and not self.pre_filter(data):
                continue

            if self.pre_transform is not None:
                data = self.pre_transform(data)

            torch.save(data, osp.join(self.processed_dir, 'data_{}.pt'.format(i)))
            i += 1

    def len(self):
        return len(self.processed_file_names)

    def get(self, idx):
        data = torch.load(osp.join(self.processed_dir, 'data_{}.pt'.format(idx)))
        return data



import os.path as osp

import torch
import torch_geometric.transforms as T
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader
from torch_geometric.utils import remove_self_loops

from pyg_extension.data.dataset import InMemoryDatasetGeo
from pyg_extension.nn.basemodel import GeneralReadOutLayer
from pyg_extension.nn.cgcnn import CrystalGraphConvNeuralNet
from pyg_extension.nn.flow_geo import LearningFlow
from pyg_extension.utils.log_model import print_log_dataset

target = 0
dim = 64


class QM9User(InMemoryDatasetGeo):

    @property
    def processed_file_names(self):
        return ['data_v3.pt']

    def process(self):
        data_list = torch.load(self.raw_paths[0])
        data_list = [Data(**data_dict) for data_dict in data_list]  # just for qm9_v3.pt data

        if self.pre_filter is not None:
            data_list = [d for d in data_list if self.pre_filter(d)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(d) for d in data_list]

        self.data, self.slices = self.collate(data_list)

        torch.save((self.data, self.slices), self.processed_paths[0])


class MyTransform(object):
    def __call__(self, data):
        # Specify target.
        data.y = data.y[:, target]
        return data


class Complete(object):
    def __call__(self, data):
        device = data.edge_index.device

        row = torch.arange(data.num_nodes, dtype=torch.long, device=device)
        col = torch.arange(data.num_nodes, dtype=torch.long, device=device)

        row = row.view(-1, 1).repeat(1, data.num_nodes).view(-1)
        col = col.repeat(data.num_nodes)
        edge_index = torch.stack([row, col], dim=0)

        edge_attr = None
        if data.edge_attr is not None:
            idx = data.edge_index[0] * data.num_nodes + data.edge_index[1]
            size = list(data.edge_attr.size())
            size[0] = data.num_nodes * data.num_nodes
            edge_attr = data.edge_attr.new_zeros(size)
            edge_attr[idx] = data.edge_attr

        edge_weight = (data.pos[row] - data.pos[col]).norm(dim=-1)

        edge_index1, edge_attr = remove_self_loops(edge_index, edge_attr)
        edge_index1, edge_weight = remove_self_loops(edge_index, edge_weight)
        data.edge_attr = edge_attr
        data.edge_weight = edge_weight
        data.edge_index = edge_index1

        data.state_attr = torch.zeros((1, 2))

        return data


path = osp.join(osp.dirname(osp.realpath(__file__)), 'qm9_v3')
transform = T.Compose([MyTransform(), Complete()])
dataset = QM9User(path, pre_transform=transform, re_process_init=False).shuffle()

# Normalize targets to mean = 0 and std = 1.

mean = dataset.data.y.mean()
std = dataset.data.y.std()
dataset.data.y = (dataset.data.y - mean) / std

# # Split datasets.
test_dataset = dataset[:1000]
val_dataset = dataset[1000:2000]
train_dataset = dataset[2000:3000]
test_loader = DataLoader(test_dataset, batch_size=128, shuffle=False)
val_loader = DataLoader(val_dataset, batch_size=128, shuffle=False)
train_loader = DataLoader(train_dataset, batch_size=128, shuffle=True)
print_log_dataset(dataset)
#
device = torch.device('cuda:1' if torch.cuda.is_available() else 'cpu')
# # model = CGGRUNet(11, 5, cutoff=5.0).to(device)
# # model = CrystalGraphConvNet(11, 5, cutoff=5.0).to(device)
# # model = CrystalGraphGCN(11, 5, cutoff=5.0).to(device)
# # model = CrystalGraphGCN2(11, 5, cutoff=5.0).to(device)
# # model = SchNet(0,0,simple_edge=True).to(device)
# # model = MEGNet(11, 5, cutoff=5.0,num_state_features=2).to(device)
# # model = SchNet(11,5,simple_edge=False).to(device)
# # model = SchNet(11,5,simple_edge=False,simple_z=False).to(device)
model = CrystalGraphConvNeuralNet(11, 4, 2, nc_node_hidden=16, nc_node_interaction=16, num_interactions=1,
                                  edge_method="gaussian", nc_edge_hidden=50, node_method="x",
                                  layer_readout=GeneralReadOutLayer([16, "mean", 1], last_layer=None),
                                  ).to(device)
optimizer = torch.optim.Adam(model.parameters(), lr=1e-4)
scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min',
                                                       min_lr=1e-3)

lf = LearningFlow(model, train_loader, validate_loader=val_loader, device="cuda:1",
                  optimizer=None, clf=False, loss_method=None, milestones=None,
                  weight_decay=0.0, checkpoint=True, scheduler=scheduler,
                  loss_threshold=0.1, print_freq=None, print_what="all")
lf.run(200)



import os
import shutil
import sys
import time
from typing import Union

import numpy as np
import torch
from sklearn import metrics
from torch.nn import Module
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.utils.data import DataLoader

from pyg_extension.utils.log_model import LogModule, HookGradientModule, make_dot_, \
    get_target_layer, AverageMeterTotal


def class_eval(prediction, target):
    """Classification."""
    prediction = np.exp(prediction.numpy())
    target = target.numpy()
    pred_label = np.argmax(prediction, axis=1)
    target_label = np.squeeze(target)
    if not target_label.shape:
        target_label = np.asarray([target_label])
    if prediction.shape[1] == 2:
        precision, recall, fscore, _ = metrics.precision_recall_fscore_support(
            target_label, pred_label, average='binary')
        auc_score = metrics.roc_auc_score(target_label, prediction[:, 1])
        accuracy = metrics.accuracy_score(target_label, pred_label)
    else:
        raise NotImplementedError
    return accuracy, precision, recall, fscore, auc_score


def mae(prediction, target):
    """
    Computes the mean absolute error between prediction and target

    Parameters
    ----------
    prediction: torch.Tensor (N, 1)
    target: torch.Tensor (N, 1)
    """
    return torch.mean(torch.abs(target - prediction))


class LearningFlow:
    """
    LearningFlow for training.
    
    Examples:

        >>> test_dataset = dataset[:1000]
        >>> val_dataset = dataset[1000:2000]
        >>> train_dataset = dataset[2000:3000]
        >>> import torch_geometric.transforms as T
        >>> train_dataset = SimpleDataset(data_train, pre_transform=T.ToSparseTensor())
        >>> test_dataset = SimpleDataset(data_test,pre_transform=T.ToSparseTensor())
        >>> val_dataset = SimpleDataset(val_data,pre_transform=T.ToSparseTensor())

        >>> train_loader = DataLoader(
        ... dataset=train_dataset,
        ... batch_size=200,
        ... shuffle=False,
        ... num_workers=0)

        >>> test_loader = ...
        >>> val_loader = ...


        >>>  model = CrystalGraphConvNet(nfeat_node=91,
        ...  nfeat_edge=3,
        ...  nfeat_state=29,
        ...  nc_node_hidden=128,
        ...  nc_node_interaction=64,
        ...  num_interactions=2,)
        >>> # model = CrystalGraphGCN(...)
        >>> # model = CrystalGraphGCN2(...)
        >>> # model = CrystalGraphGAT(...)
        >>> # model = SchNet(...)
        >>> # model = MEGNet(...)
        >>> # model = SchNet(...)


        >>> optimizer = torch.optim.Adam(model.parameters(), lr=0.01)
        >>> scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.7, patience=2,...min_lr=0.001)
        >>> device = torch.device('cuda:1' if torch.cuda.is_available() else 'cpu')

        >>> lf= LearningFlow(model, train_loader, validate_loader=val_loader, device= "cuda:1",
        ... optimizer=None, clf= False, loss_method=None, learning_rate = 1e-3,
        ... weight_decay= 0.01, checkpoint=True, scheduler=scheduler,
        ... loss_threshold= 0.1, print_freq= None, print_what="all")

        >>> lf.run(50)

    where the dataset could from StructureToData or :class:`featurebox.test_featurizers.base_graph_geo.StructureGraphGEO`
    """

    def __init__(self, model: Module, train_loader: DataLoader, validate_loader: DataLoader, device="cpu",
                 optimizer=None, clf: bool = False, loss_method=None, learning_rate: float = 1e-3,
                 weight_decay: float = 0.0, checkpoint=True, scheduler=None, debug: str = "", target_layers=(),
                 loss_threshold: float = 0.1, print_freq: Union[int, str, None] = 10, print_what="all",
                 process_label=None):
        """

        Parameters
        ----------
        model: module
        train_loader: DataLoader
        validate_loader: DataLoader
        device:str
            such as "cuda:0","cpu"
        optimizer:torch.Optimizer
        clf:bool
            under exploit......
        loss_method:torch._Loss
            see more in torch
        learning_rate:float
            see more in torch
        weight_decay:float
            see more in torch
        checkpoint:bool
            save checkpoint or not.
        loss_threshold:
            see more in torch
        print_freq:int
            print frequency
        print_what:str
            "all","train","test" log.
        scheduler:
            scheduler, see more in torch
        process_label:Callable
            function to get y/label.
        """

        self.train_loader = train_loader
        self.test_loader = validate_loader

        device = torch.device(device)

        self.device = device
        self.model = model
        self.model.to(device)
        self.clf = clf
        self.loss_method = loss_method
        self.optimizer = optimizer
        self.checkpoint = checkpoint
        self.weight_log = LogModule()

        self.debug = debug

        self.train_batch_number = len(self.train_loader)

        if print_freq is None:
            self.print_freq = self.train_batch_number
        else:
            self.print_freq = self.train_batch_number if not isinstance(print_freq, int) else print_freq

        if self.optimizer is None or self.optimizer == "Adam":
            self.optimizer = torch.optim.Adam(self.model.parameters(), lr=learning_rate, weight_decay=weight_decay)
        elif self.optimizer == "SGD":
            self.optimizer = torch.optim.SGD(self.model.parameters(), lr=learning_rate, weight_decay=weight_decay
                                             )
            # L2 regularization
        else:
            self.optimizer = optimizer(self.model.parameters(), lr=learning_rate, weight_decay=weight_decay)

        if self.loss_method is None:
            if self.clf is True:
                self.loss_method = torch.nn.CrossEntropyLoss()
            elif self.clf == "multi_label":
                self.loss_method = torch.nn.L1Loss()
                # 主要是用来判定实际的输出与期望的输出的接近程度 MAE: 1/N |y_pred-y| y 为多列
            else:
                self.loss_method = torch.nn.MSELoss()

        else:
            self.loss_method = loss_method
        if scheduler is None:
            self.scheduler = ReduceLROnPlateau(self.optimizer, mode='min', factor=0.5, patience=7, min_lr=0.00001)
        else:
            self.optimizer = scheduler.optimizer
            self.scheduler = scheduler
        self.best_error = 1000000.0
        self.threshold = loss_threshold
        # *.pth.tar or str
        self.run_train = self.run
        self.fit = self.run_train
        self.print_what = print_what
        self.forward_hook_list = []
        func = lambda x: x
        self._process_label = func if process_label is None else process_label
        self.target_layers = target_layers

        if self.clf:
            self.meters = AverageMeterTotal("Time", "losses", "accuracies", "precisions", "recalls", "fscores",
                                            "auc_scores")
        else:
            self.meters = AverageMeterTotal("Time", "losses", "mae_errors")

    def my_debug(self):
        # debug
        if "loop" in self.debug:
            self.weight_log.stats_loop()

        if "single" in self.debug:
            self.model.train()
            for m, data in enumerate(self.train_loader):  # just record for once,
                data = data.to(self.device)
                self.model(data)
                self.weight_log.get_weight(self.model, append=False)
                break
            self.weight_log.stats_single()

        if "graphviz" in self.debug:

            self.model.train()
            for m, data in enumerate(self.train_loader):  # just record for once,
                data = data.to(self.device)
                y_pred = self.model(data)
                vis_graph = make_dot_(y_pred, params=dict(list(self.model.named_parameters())))
                vis_graph.render(format="pdf",view=False)
                break

        if "hook" in self.debug:
            # which layer to hook.
            self.target_layers = get_target_layer(self.model, self.target_layers)
            self.gradient_log = HookGradientModule(self.target_layers)
            self.model.train()
            for m, data in enumerate(self.train_loader):  # just record for once,
                data = data.to(self.device)
                batch_y = self.process_label(data.y)
                self.model.zero_grad()
                y_pred = self.model(data)
                lossi = self.loss_method(y_pred, batch_y)
                lossi.backward(retain_graph=True)
                # lossi.backward()
                break
            self.gradient_log.stats_single()

    def process_label(self, y):
        return self._process_label(y)

    @staticmethod
    def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
        torch.save(state, filename)
        if is_best:
            shutil.copyfile(filename, 'model_best.pth.tar')

    def run(self, epoch=50, warm_start=False):
        """
        run loop.

        Parameters
        ----------
        epoch:int
            epoch.
        warm_start: str, False
            The name of resume file, 'checkpoint.pth.tar' or 'model_best.pth.tar'
            If warm_start, try to resume from local disk.
        """

        resume = warm_start if warm_start is not False else None
        start_epoch = 0
        if resume:
            if os.path.isfile(resume):
                print("=> loading checkpoint '{}'".format(resume))
                checkpoint = torch.load(resume)
                start_epoch = checkpoint['epoch']
                self.threshold = checkpoint['best_error']
                self.model.load_state_dict(checkpoint['state_dict'])
                self.optimizer.load_state_dict(checkpoint['optimizer'])
                print("=> loaded checkpoint '{}' (epoch {})"
                      .format(resume, checkpoint['epoch']))
            else:
                print("=> no checkpoint found at '{}'".format(resume))

        if start_epoch > 0:
            epoch += start_epoch
            print("Try to run start from 'resumed epoch' {} to 'epoch' {}".format(start_epoch, epoch))

        self.model.to(self.device)

        for epochi in range(start_epoch, epoch):

            self._train(epochi)

            if "loop" in self.debug in self.debug:  # record each epochi
                self.weight_log.get_weight(self.model)

            score = self._validate(epochi)

            if score != score:
                print('Exit due to NaN')
                sys.exit(1)
            try:
                self.scheduler.step(score)
            except TypeError:
                self.scheduler.step()

            is_best = score < self.best_error
            self.best_error = min(score, self.best_error)

            if self.checkpoint:
                self.save_checkpoint({
                    'epoch': epochi + 1,
                    'state_dict': self.model.state_dict(),
                    'best_error': self.threshold,
                    'optimizer': self.optimizer.state_dict(),
                }, is_best)

            if score <= self.threshold:
                print("Up to requirements and early termination in epoch ({})".format(epochi))
                break

        self.my_debug()

    def _train(self, epochi):
        self.model.train()
        self.meters.reset()

        for m, data in enumerate(self.train_loader):
            point = time.time()
            data = data.to(self.device)
            batch_y = self.process_label(data.y)

            self.optimizer.zero_grad()

            y_pred = self.model(data)

            try:
                lossi = self.loss_method(y_pred, batch_y)
            except TypeError:
                target = batch_y.sign()
                lossi = self.loss_method(y_pred, batch_y, target)

            lossi.backward()

            self.optimizer.step()

            # ###########

            batch_size = batch_y.size(0)
            self.meters['Time'].record(time.time() - point)
            self.meters['losses'].record(float(lossi.cpu().item()), batch_size)
            if self.clf is False:
                mae_error = mae(y_pred.data.cpu(), batch_y.cpu())
                self.meters['mae_errors'].record(mae_error, batch_size)
            else:
                accuracy, precision, recall, fscore, auc_score = \
                    class_eval(y_pred.data.cpu(), batch_y.cpu())
                self.meters['accuracies'].record(accuracy, batch_size)
                self.meters['precisions'].record(precision, batch_size)
                self.meters['recalls'].record(recall, batch_size)
                self.meters['fscores'].record(fscore, batch_size)
                self.meters['auc_scores'].record(auc_score, batch_size)

            if self.print_freq != self.train_batch_number and m % self.print_freq == 0 \
                    and self.print_what in ["all", "train"]:
                print('Train: [{0}][{1}/{2}] {3}'.format(epochi, m, len(self.train_loader), self.meters.text()))

        if self.print_freq == self.train_batch_number:
            print('Train: [{}] {}'.format(epochi, self.meters.text()))

    def _validate(self, epochi):
        self.model.eval()
        self.meters.reset()

        for m, data in enumerate(self.test_loader):
            point = time.time()

            data = data.to(self.device)
            batch_y = self.process_label(data.y)
            y_pred = self.model(data)

            try:
                lossi = self.loss_method(y_pred, batch_y)
            except TypeError:
                target = batch_y.sign()
                lossi = self.loss_method(y_pred, batch_y, target)

            # ###########

            batch_size = batch_y.size(0)
            self.meters['Time'].record(time.time() - point)
            self.meters['losses'].record(float(lossi.cpu().item()), batch_size)
            if self.clf is False:
                mae_error = mae(y_pred.data.cpu(), batch_y.cpu())
                self.meters['mae_errors'].record(mae_error, batch_size)
            else:
                accuracy, precision, recall, fscore, auc_score = \
                    class_eval(y_pred.data.cpu(), batch_y.cpu())
                self.meters['accuracies'].record(accuracy, batch_size)
                self.meters['precisions'].record(precision, batch_size)
                self.meters['recalls'].record(recall, batch_size)
                self.meters['fscores'].record(fscore, batch_size)
                self.meters['auc_scores'].record(auc_score, batch_size)

            if self.print_freq != self.train_batch_number and m % self.print_freq == 0 \
                    and self.print_what in ["all", "train"]:
                print('Test: [{0}][{1}/{2}] {3}'.format(epochi, m, len(self.train_loader), self.meters.text()))

        if self.print_freq == self.train_batch_number:
            print('Test: [{}] {}'.format(epochi, self.meters.text()))

        if self.clf is False:
            return self.meters['mae_errors'].avg
        else:
            return self.meters['auc_scores'].avg

    def score(self, predict_loader):
        """Return MAE score."""
        y_pre, y_true = self.predict(predict_loader)
        return float(mae(y_pre, y_true))

    def predict(self, predict_loader: DataLoader, return_y_true=False, add_hook=False, hook_layer=None):
        """
        Just predict by model,and add one forward hook to get processing output.

        Parameters
        ----------
        predict_loader:DataLoader
            MGEDataLoader, the target_y could be ``None``.
        return_y_true:bool
            if return_y_true, return (y_preds, y_true)
        add_hook:bool
        hook_layer:
            layer of model. could get layers by ``get_target_layer`` or ``get_layers_name_with_weight``.

        Returns
        -------
        y_pred:tensor
        y_true:tensor
            if return_y_true

        """

        self.model.eval()
        self.model.to(self.device)

        ############

        handles = []
        if add_hook:
            try:
                self.forward_hook_list = []

                def for_hook(module, input, output):
                    self.forward_hook_list.append(output.detach().cpu())

                handles.append(hook_layer.register_forward_hook(for_hook))
            except BaseException as e:
                print(e)
                raise AttributeError("use ``hook_layer_name`` to defined the hook layer.")
        y_preds = []
        y_true = []
        for data in predict_loader:
            data = data.to(self.device)
            y_preds.append(self.model(data).detach().cpu())
            if hasattr(data, "y"):
                y_true.append(self.process_label(data.y.detach().cpu()))

        if add_hook:
            [i.remove() for i in handles]  # del

        if return_y_true and y_true != []:
            return torch.cat(y_preds), torch.cat(y_true)
        else:
            return torch.cat(y_preds)



#!/usr/bin/python3.7
# -*- coding: utf-8 -*-


from os import path

from setuptools import setup, find_packages

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(

    name='pyg_extension',
    version='0.0.04',
    description='Crystal extension for pytorch_geometric',
    install_requires=['torch-geometric', "torch", "pymatgen", "ase"],
    include_package_data=True,
    author='wangchangxin',
    author_email='986798607@qq.com',
    python_requires='>=3.6',
    maintainer='wangchangxin',
    platforms=[
        "Windows",
        "Unix",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],

    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests", "test"], ),
    long_description=long_description,
    long_description_content_type='text/markdown'
)
